import base64
import importlib.metadata
import struct

__version__ = importlib.metadata.version("django-pgcrypto")
__version_info__ = tuple(
    int(num) if num.isdigit() else num for num in __version__.split(".")
)

CRC24_INIT = 0xB704CE
CRC24_POLY = 0x1864CFB


class BadChecksumError(Exception):
    pass


def ord_safe(ch):
    if isinstance(ch, int):
        return ch
    return ord(ch)


def crc24(data):
    crc = CRC24_INIT
    for byte in data:
        crc ^= ord_safe(byte) << 16
        for _i in range(8):
            crc <<= 1
            if crc & 0x1000000:
                crc ^= CRC24_POLY
    return crc & 0xFFFFFF


def armor(data, versioned=True):
    """
    Returns a string in ASCII Armor format, for the given binary data. The
    output of this is compatiple with pgcrypto's armor/dearmor functions.
    """
    template = (
        "-----BEGIN PGP MESSAGE-----\n"
        "%(headers)s%(body)s\n=%(crc)s\n"
        "-----END PGP MESSAGE-----"
    )
    body = base64.b64encode(data)
    # The 24-bit CRC should be in big-endian, strip off the first byte (it's already
    # masked in crc24).
    crc = base64.b64encode(struct.pack(">L", crc24(data))[1:])
    return template % {
        "headers": "Version: django-pgcrypto %s\n\n" % __version__
        if versioned
        else "\n",
        "body": body.decode("ascii"),
        "crc": crc.decode("ascii"),
    }


def dearmor(text, verify=True):
    """
    Given a string in ASCII Armor format, returns the decoded binary data.
    If verify=True (the default), the CRC is decoded and checked against that
    of the decoded data, otherwise it is ignored. If the checksum does not
    match, a BadChecksumError exception is raised.
    """
    lines = text.strip().split("\n")
    data_lines = []
    check_data = None
    started = False
    in_body = False
    for line in lines:
        if line.startswith("-----BEGIN"):
            started = True
        elif line.startswith("-----END"):
            break
        elif started:
            if in_body:
                if line.startswith("="):
                    # Once we get the checksum data, we're done.
                    check_data = line[1:5].encode("ascii")
                    break
                else:
                    # This is part of the base64-encoded data.
                    data_lines.append(line)
            else:
                if line.strip():
                    # This is a header line, which we basically ignore for now.
                    pass
                else:
                    # The data starts after an empty line.
                    in_body = True
    b64_str = "".join(data_lines)
    # Python 3's b64decode expects bytes, not a string. We know base64 is ASCII, though.
    data = base64.b64decode(b64_str.encode("ascii"))
    if verify and check_data:
        # The 24-bit CRC is in big-endian, so we add a null byte to the beginning.
        crc = struct.unpack(">L", b"\0" + base64.b64decode(check_data))[0]
        if crc != crc24(data):
            raise BadChecksumError()
    return data


def unpad(text, block_size):
    """
    Takes the last character of the text, and if it is less than the block_size,
    assumes the text is padded, and removes any trailing zeros or bytes with the
    value of the pad character. See http://www.di-mgt.com.au/cryptopad.html for
    more information (methods 1, 3, and 4).
    """
    end = len(text)
    if end == 0:
        return text
    padch = ord_safe(text[end - 1])
    if padch > block_size:
        # If the last byte value is larger than the block size, it's not padded.
        return text
    while end > 0 and ord_safe(text[end - 1]) in (0, padch):
        end -= 1
    return text[:end]


def pad(text, block_size, zero=False):
    """
    Given a text string and a block size, pads the text with bytes of the same value
    as the number of padding bytes. This is the recommended method, and the one used
    by pgcrypto. See http://www.di-mgt.com.au/cryptopad.html for more information.
    """
    num = block_size - (len(text) % block_size)
    ch = b"\0" if zero else chr(num).encode("latin-1")
    return text + (ch * num)


def aes_pad_key(key):
    """
    AES keys must be either 16, 24, or 32 bytes long. If a key is provided that is not
    one of these lengths, pad it with zeroes (this is what pgcrypto does).
    """
    if len(key) in (16, 24, 32):
        return key
    if len(key) < 16:
        return pad(key, 16, zero=True)
    elif len(key) < 24:
        return pad(key, 24, zero=True)
    elif len(key) < 32:
        return pad(key, 32, zero=True)
    else:
        return key[:32]



from django.conf import settings
from django.db.models import Func

from .base import aes_pad_key


class CryptoFunc(Func):
    def __init__(self, *args, **kwargs):
        self.params = {
            "cipher_name": kwargs.pop("cipher", None),
            "cipher_key": kwargs.pop("key", None),
            "charset": kwargs.pop("charset", None),
        }

        super().__init__(*args, **kwargs)

    def get_params(self):
        cipher_name = self.params["cipher_name"] or getattr(
            self.field,
            "cipher_name",
            getattr(settings, "PGCRYPTO_DEFAULT_CIPHER", "aes"),
        )
        cipher_key = self.params["cipher_key"] or getattr(
            self.field,
            "cipher_key",
            getattr(settings, "PGCRYPTO_DEFAULT_KEY", settings.SECRET_KEY),
        )
        charset = self.params["charset"] or getattr(self.field, "charset", "utf-8")

        if cipher_name == "aes":
            # If this is from self.field, it's already padded, but this is still safe
            # to call again.
            cipher_key = aes_pad_key(cipher_key)

        return cipher_name, cipher_key, charset


class Encrypt(CryptoFunc):
    function = "encrypt"
    template = (
        "armor(%(function)s(convert_to(nullif(%(expressions)s, ''), %%s), %%s, %%s))"
    )

    def as_sql(self, *args, **extra_context):
        sql, params = super().as_sql(*args, **extra_context)
        cipher_name, cipher_key, charset = self.get_params()
        return sql, (*params, charset, cipher_key, cipher_name)


class Decrypt(CryptoFunc):
    function = "decrypt"
    template = "convert_from(%(function)s(dearmor(nullif(%(expressions)s, '')), %%s, %%s), %%s)"

    def as_sql(self, *args, **extra_context):
        sql, params = super().as_sql(*args, **extra_context)
        cipher_name, cipher_key, charset = self.get_params()
        return sql, (*params, cipher_key, cipher_name, charset)



# django-pgcrypto
# Dan Watson
#
# A pure python implementation of ASCII Armor, along with various
# padding and unpadding functions, all compatible with pgcrypto.
#
# Additionally, this module defines Django fields that automatically
# encrypt and armor (and decrypt and dearmor) values for storage
# in text fields. Values stored using these fields may be read by
# pgcrypto using decrypt(dearmor(col),...), and values stored by
# pgcrypto using armor(encrypt(col,...)) may be read by these fields.
#
# See http://www.ietf.org/rfc/rfc2440.txt for ASCII Armor specs.

from .base import __version__, __version_info__, aes_pad_key, armor, dearmor, pad, unpad

__all__ = [
    "__version__",
    "__version_info__",
    "aes_pad_key",
    "armor",
    "dearmor",
    "pad",
    "unpad",
]

try:
    import django  # noqa: F401

    has_django = True
except ImportError:
    has_django = False

if has_django:
    from .fields import (
        EncryptedCharField,
        EncryptedDateField,
        EncryptedDateTimeField,
        EncryptedDecimalField,
        EncryptedEmailField,
        EncryptedIntegerField,
        EncryptedTextField,
    )
    from .functions import Decrypt, Encrypt

    __all__ += [
        "EncryptedCharField",
        "EncryptedDateField",
        "EncryptedDateTimeField",
        "EncryptedDecimalField",
        "EncryptedEmailField",
        "EncryptedIntegerField",
        "EncryptedTextField",
        "Encrypt",
        "Decrypt",
    ]



import datetime
import decimal

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from django import forms
from django.conf import settings
from django.core import validators
from django.db import models
from django.db.models.lookups import FieldGetDbPrepValueIterableMixin, Lookup
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from .base import aes_pad_key, armor, dearmor, pad, unpad


class BaseEncryptedField(models.Field):
    field_cast = ""
    coalesce = None

    def __init__(self, *args, **kwargs):
        self.cipher_name = kwargs.pop(
            "cipher", getattr(settings, "PGCRYPTO_DEFAULT_CIPHER", "aes")
        ).lower()
        if self.cipher_name != "aes":
            raise ValueError("Cipher must be `aes` (got `{}`)".format(self.cipher_name))
        self.cipher_key = kwargs.pop(
            "key", getattr(settings, "PGCRYPTO_DEFAULT_KEY", settings.SECRET_KEY)
        )
        self.charset = kwargs.pop("charset", "utf-8")
        if isinstance(self.cipher_key, str):
            self.cipher_key = self.cipher_key.encode(self.charset)
        if self.cipher_name == "aes":
            self.cipher_key = aes_pad_key(self.cipher_key)
        self.check_armor = kwargs.pop("check_armor", True)
        self.versioned = kwargs.pop("versioned", False)
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "TextField"

    def deconstruct(self):
        """
        Deconstruct the field for Django 1.7+ migrations.
        """
        name, path, args, kwargs = super().deconstruct()
        kwargs.update(
            {
                "cipher": self.cipher_name,
                "charset": self.charset,
                "check_armor": self.check_armor,
                "versioned": self.versioned,
            }
        )
        return name, path, args, kwargs

    @property
    def algorithm(self):
        return {"aes": algorithms.AES}[self.cipher_name]

    @property
    def block_size(self):
        return self.algorithm.block_size // 8

    def get_cipher(self):
        """
        Return a new Cipher object for each time we want to encrypt/decrypt. This is
        because pgcrypto expects a zeroed block for IV (initial value), but the IV on
        the cipher object is cumulatively updated each time encrypt/decrypt is called.
        """
        return Cipher(
            self.algorithm(self.cipher_key),
            modes.CBC(b"\0" * self.block_size),
            backend=default_backend(),
        )

    def encrypt(self, data):
        context = self.get_cipher().encryptor()
        return context.update(data) + context.finalize()

    def decrypt(self, data):
        context = self.get_cipher().decryptor()
        return context.update(data) + context.finalize()

    def is_encrypted(self, value):
        """
        Returns whether the given value is encrypted (and armored) or not.
        """
        return isinstance(value, str) and value.startswith("-----BEGIN")

    def to_python(self, value):
        if self.is_encrypted(value):
            # If we have an encrypted (armored, really) value, do the following when
            # accessing it as a python value:
            #    1. De-armor the value to get an encrypted bytestring.
            #    2. Decrypt the bytestring using the specified cipher.
            #    3. Unpad the bytestring using the cipher's block size.
            #    4. Decode to a unicode string using the specified charset.
            return unpad(
                self.decrypt(dearmor(value, verify=self.check_armor)), self.block_size
            ).decode(self.charset)
        return value

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def get_db_prep_save(self, value, connection):
        if hasattr(value, "as_sql"):
            # If the value is a query expression do not encrypt it, it will circle back to this function to
            # encrypt the value in the Val() expression within the query.
            return value

        if value and not self.is_encrypted(value):
            # If we have a value and it's not encrypted, do the following before storing
            # in the database:
            #    1. Convert it to a unicode string (by calling unicode).
            #    2. Encode the unicode string according to the specified charset.
            #    3. Pad the bytestring for encryption, using the cipher's block size.
            #    4. Encrypt the padded bytestring using the specified cipher.
            #    5. Armor the encrypted bytestring for storage in the text field.
            return armor(
                self.encrypt(
                    pad(force_str(value).encode(self.charset), self.block_size)
                ),
                versioned=self.versioned,
            )
        return value


class EncryptedTextField(BaseEncryptedField):
    description = _("Text")
    coalesce = "''"

    def formfield(self, **kwargs):
        defaults = {"widget": forms.Textarea}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class EncryptedCharField(BaseEncryptedField):
    description = _("String")
    coalesce = "''"

    def __init__(self, *args, **kwargs):
        # We don't want to restrict the max_length of an EncryptedCharField
        # because of the extra characters from encryption, but we'd like
        # to use the same interface as CharField
        kwargs.pop("max_length", None)
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        defaults = {"widget": forms.TextInput}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class EncryptedIntegerField(BaseEncryptedField):
    description = _("Integer")
    field_cast = "::integer"

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.IntegerField}
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def to_python(self, value):
        if value:
            return int(super().to_python(value))
        return value


class EncryptedDecimalField(BaseEncryptedField):
    description = _("Decimal number")
    field_cast = "::numeric"

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.DecimalField}
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def to_python(self, value):
        if value:
            return decimal.Decimal(super().to_python(value))
        return value


class EncryptedDateField(BaseEncryptedField):
    description = _("Date (without time)")
    field_cast = "::date"

    def __init__(
        self,
        verbose_name=None,
        name=None,
        auto_now=False,
        auto_now_add=False,
        **kwargs,
    ):
        self.auto_now, self.auto_now_add = auto_now, auto_now_add
        if auto_now or auto_now_add:
            kwargs["editable"] = False
            kwargs["blank"] = True
        super().__init__(verbose_name, name, **kwargs)

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.DateField}
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def to_python(self, value):
        if value in self.empty_values:
            return None
        unencrypted_value = super().to_python(value)
        return self._parse_value(unencrypted_value)

    def value_to_string(self, obj):
        val = self.value_from_object(obj)
        return "" if val is None else val.isoformat()

    def pre_save(self, model_instance, add):
        if self.auto_now or (self.auto_now_add and add):
            value = self._get_auto_now_value()
            setattr(model_instance, self.attname, value)
            return value
        else:
            return super().pre_save(model_instance, add)

    def _parse_value(self, value):
        return models.DateField().to_python(value)

    def _get_auto_now_value(self):
        return datetime.date.today()


class EncryptedDateTimeField(EncryptedDateField):
    description = _("Date (with time)")
    field_cast = "timestamp with time zone"

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.DateTimeField}
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def _parse_value(self, value):
        return models.DateTimeField().to_python(value)

    def _get_auto_now_value(self):
        return timezone.now()


class EncryptedEmailField(BaseEncryptedField):
    default_validators = [validators.validate_email]
    description = _("Email address")
    coalesce = "''"

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.EmailField}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class EncryptedLookup(Lookup):
    patterns = {
        "contains": "%%%s%%",
        "icontains": "%%%s%%",
        "startswith": "%s%%",
        "istartswith": "%s%%",
        "endswith": "%%%s",
        "iendswith": "%%%s",
    }

    def as_postgresql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        if self.lookup_name == "in":
            rhs = "IN (%s)" % ", ".join(rhs)
        else:
            rhs = connection.operators[self.lookup_name] % rhs
        if self.lookup_name == "exact" and rhs_params and rhs_params[0] == "":
            # Special case when looking for blank values, don't try to dearmor/decrypt.
            return "%s %s" % (lhs, rhs), (*lhs_params, *rhs_params)
        field_sql = (
            "convert_from(decrypt(dearmor(nullif(%s, '')), %%s, '%s'), 'utf-8')"
            % (lhs, self.lhs.output_field.cipher_name)
        )
        if self.lhs.output_field.coalesce:
            field_sql = (
                "coalesce(" + field_sql + ", " + self.lhs.output_field.coalesce + ")"
            )
        field_internal_type = self.lhs.output_field.get_internal_type()
        field_sql = (
            connection.ops.lookup_cast(self.lookup_name, field_internal_type)
            % field_sql
        )

        if (
            self.lookup_name in self.patterns
            and self.rhs_is_direct_value()
            and rhs_params
            and not self.bilateral_transforms
        ):
            pat = self.patterns[self.lookup_name] % connection.ops.prep_for_like_query(
                rhs_params[0]
            )
            rhs_params = (pat, *rhs_params[1:])

        return (
            "%s%s %s"
            % (
                field_sql,
                self.lhs.output_field.field_cast,
                rhs,
            ),
            (
                *lhs_params,
                self.lhs.output_field.cipher_key,
                *rhs_params,
            ),
        )


class EncryptedInLookup(FieldGetDbPrepValueIterableMixin, EncryptedLookup):
    lookup_name = "in"


for lookup_name in (
    "exact",
    "iexact",
    "gt",
    "gte",
    "lt",
    "lte",
    "contains",
    "icontains",
    "startswith",
    "istartswith",
    "endswith",
    "iendswith",
):
    class_name = "EncryptedLookup_%s" % lookup_name
    lookup_class = type(class_name, (EncryptedLookup,), {"lookup_name": lookup_name})
    BaseEncryptedField.register_lookup(lookup_class)

BaseEncryptedField.register_lookup(EncryptedInLookup)
